#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>                           /* for gettimeofday           */
#include <string.h>
#include <mmintrin.h>

/* #define FILE_HEADER_SIZE         1078           Windows BMP header size.   */
#define FILE_HEADER_SIZE         54           /* BMP header size.   */
#define MMX2                     2
#define MMX1                     1
#define ASM                      3
#define C                        0

#define INTRINSIC 1

int mmx_init(void);
void parse_args(int, char **);
void mmx_brighten(unsigned char *, long);
void pure_c_brighten(unsigned char *, long);
void asm_brighten(unsigned char *, long);
void usage(void);

double teff1, teff2;
long i, t1, t2, tm1, tm2;
struct timeval tv;
char input_file_name[100], output_file_name[100];
unsigned char flag, brt;

const char *bmpbright_usage = "brightening <inputfile.bmp> <outputfile.bmp> <brightness> {c | mmx | asm}";

int mmx_init(void)
{
	int mmx_available;
	__asm__("movl $1, %%eax                      # get CPU version information\n\t"
	        "cpuid\n\t"
	        "andl $0x800000, %%edx               # check bit 23 of edx\n\t"
	        "movl %%edx, %0\n\t"
        : "=r"(mmx_available)
			        :
			        : "%eax", "%ebx", "%ecx", "%edx"
			       );
	return(mmx_available);
}


void mmx_brighten(unsigned char *array, long length)
{
	/* Hier noch evtl. benoetigte Vorbereitungen */
	unsigned char brt8[8];

#if INTRINSIC
	__m64 _mm_brt8;
	__m64* _mm_array;
	__m64* _mm_end;
#endif

	for (i = 0; i < 8; i++) {
		brt8[i] = brt;
	}

	printf("Finished. Press <Enter> to perform MMX brightening on array...\n");
	i = getchar();

	gettimeofday(&tv, NULL);
	t1 = tv.tv_sec;
	tm1 = tv.tv_usec;

#if INTRINSIC
	memcpy(&_mm_brt8, &brt8, sizeof(__m64));

	_mm_end = (__m64*)(&array[length]);
	for (_mm_array = (__m64*)array; _mm_array < _mm_end; _mm_array++) {
		*_mm_array = _mm_adds_pu8(*_mm_array, _mm_brt8);
	}
	_mm_empty();
#else
	for (i = 0; i < length; i += 8) {
		/* Hier Assembler-Stueck fuer MMX-Brightening */
		__asm__(
		        "movq (%1), %%mm0\n\t"
		        "paddusb (%0), %%mm0\n\t"
		        "movq %%mm0, (%1)\n\t"
		        "emms\n\t"
	        : /* keine Ausgabeparameter */
	        : "r"(brt8), "r"(array + i)
		        /* Argument %0 = Brightening-Wert */
		        /* Argument %1 = Adresse des zu behandelnden Bytes */
				        : "mm0" /* mm0 wird veraendert */
				);
	}
#endif

	gettimeofday(&tv, NULL);
	t2 = tv.tv_sec;
	tm2 = tv.tv_usec;
	teff1 = t1 * 1e6 + tm1;
	teff2 = t2 * 1e6 + tm2;
	printf("Finished.\n");
	printf("%f, %f\n", teff1, teff2);
	printf("Time for brightening using MMX in usec.: %f\n", teff2 - teff1);
}

void pure_c_brighten(unsigned char *array, long length)
{
	unsigned int temp;

	printf("Finished. Press <Enter> to perform C brightening on array...\n");
	i = getchar();

	gettimeofday(&tv, NULL);
	t1 = tv.tv_sec;
	tm1 = tv.tv_usec;

	for (i = 0; i < length; i++) {
		temp = array[i] + brt;

		if (temp > 255)
			temp = 255;

		array[i] = temp;
	}

	gettimeofday(&tv, NULL);
	t2 = tv.tv_sec;
	tm2 = tv.tv_usec;
	teff1 = t1 * 1e6 + tm1;
	teff2 = t2 * 1e6 + tm2;
	printf("Finished.\n");
	printf("Time for brightening using C in usec.: %f\n", teff2 - teff1);
}

void asm_brighten(unsigned char *array, long length)
{

	printf("Finished. Press <Enter> to perform ASM brightening on array...\n");
	i = getchar();

	gettimeofday(&tv, NULL);
	t1 = tv.tv_sec;
	tm1 = tv.tv_usec;

	for (i = 0; i < length; i++) {
		__asm__(
		        "movb (%1), %%al   # Byte laden (von Adresse spezifiziert durch Arg. %1)\n\t"
		        "addb %0, %%al     # Wert (uebergeben in Argument %0) hinzuaddieren\n\t"
		        "jnc .asmexit      # Falls Carry nicht gesetzt fertig (Ergebnis <= 255)\n\t"
		        "movb $255, %%al   # Ergebnis war > 255, also auf Maximalwert setzen\n\t"
		        ".asmexit:\n\t"
		        "movb %%al, (%1)   # veraendertes Byte zurueckschreiben\n\t"
	        : /* keine Ausgabeparameter */
	        : "r"(brt), "r"(array + i)
		        /* Argument %0 = Brightening-Wert */
		        /* Argument %1 = Adresse des zu behandelnden Bytes */
				        : "al" /* al wird veraendert */
				);
	}

	gettimeofday(&tv, NULL);
	t2 = tv.tv_sec;
	tm2 = tv.tv_usec;
	teff1 = t1 * 1e6 + tm1;
	teff2 = t2 * 1e6 + tm2;
	printf("Finished.\n");
	printf("Time for brightening using ASM in usec.: %f\n", teff2 - teff1);
}

void usage(void)
{
	printf("Command line parse error!\n\nusage: %s\n", bmpbright_usage);
	exit(-1);
}

void parse_args(int argc, char **argv)
{
	if (argc < 4) {
		usage();
	}

	argc--;

	if (!(strcmp(argv[argc], "mmx"))) {
		printf("Mode: MMX\n");
		flag = MMX1;
	} else if (!(strcmp(argv[argc], "c"))) {
		printf("Mode: pure C\n");
		flag = C;
	} else if (!(strcmp(argv[argc], "asm"))) {
		printf("Mode: Assembler (no MMX)\n");
		flag = ASM;
	} else {
		usage();
	}

	argc--;
	brt = (unsigned char)atoi(argv[argc]);
	printf("Brightness constant: %d\n", brt);
	argc--;
	strcpy(output_file_name, argv[argc]);
	printf("Output file: %s\n", output_file_name);
	argc--;
	strcpy(input_file_name, argv[argc]);
	printf("Input file: %s\n", input_file_name);
}

int main(int argc, char **argv)
{
	FILE *infile, *outfile;
	unsigned char * bitmap;
	unsigned char header[FILE_HEADER_SIZE];
	size_t bitmap_size, adjusted_size;

	if (!(mmx_init() == 0x800000)) {
		printf("Sorry, your processor does not support MMX technology.\n");
		printf("This program only runs on Pentium MMX class processors\n\n");
		exit(-1);
	}

	parse_args(argc, argv);

	printf("Loading image into memory...\n");

	infile = fopen(input_file_name, "rb");

	fseek(infile, 0L, SEEK_END);
	bitmap_size = ftell(infile) - FILE_HEADER_SIZE;        /* calculate the size of the bitmap */

	adjusted_size = 8 * (bitmap_size / 8) + 8;             /* the next number which is a multiple of 8 */
	bitmap = (unsigned char *) malloc(adjusted_size * sizeof(char));

	rewind(infile);                                       /* Read the header and the bitmap from the input file */
	fread(header, sizeof(unsigned char), FILE_HEADER_SIZE, infile);
	fread(bitmap, sizeof(unsigned char), bitmap_size, infile);
	fclose(infile);

	if (flag == MMX1) {
		mmx_brighten(bitmap, adjusted_size);
	} else if (flag == ASM) {
		asm_brighten(bitmap, adjusted_size);
	} else {
		pure_c_brighten(bitmap, adjusted_size);
	}

	outfile = fopen(output_file_name, "wb");
	fwrite(header, sizeof(char), FILE_HEADER_SIZE, outfile);
	fwrite(bitmap, sizeof(char), bitmap_size, outfile);
	fclose(outfile);

	free(bitmap);

	return(0);
}
